# vim: set filetype=python ts=4 sw=4
# -*- coding: utf-8 -*-
"""This module is responsible for initialisation of global variables."""
from __future__ import absolute_import, division, print_function, unicode_literals

from os.path import expanduser
import sys

from future import standard_library

standard_library.install_aliases()


config_dir = expanduser("~") + "/.aws"
config_file = config_dir + "/okta_auth"
aws_config_file = config_dir + "/config"
aws_shared_credentials_file = config_dir + "/credentials"
aws_output = "json"
aws_profile = None
aws_region = "us-east-1"
encoding = sys.stdin.encoding
mfa_method = None
mfa_response = None
okta_aws_app_url = None
okta_status_dict = {
    "E0000004": "Authentication failed",
    "E0000047": "API call exceeded rate limit due to too many requests",
    "PASSWORD_EXPIRED": "Your password is expired",
    "LOCKED_OUT": "Your account is locked out",
}
okta_org = None
okta_password = ""
okta_profile = "default"
okta_username = ""
role_arn = None



# vim: set filetype=python ts=4 sw=4
# -*- coding: utf-8 -*-
"""Helper module for AWS and Okta configuration, management and data flow."""
from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
from builtins import (  # noqa: F401
    ascii,
    bytes,
    chr,
    dict,
    filter,
    hex,
    input,
    int,
    list,
    map,
    next,
    object,
    oct,
    open,
    pow,
    range,
    round,
    str,
    super,
    zip,
)
import codecs
import configparser
import getpass
import json
import logging
import os
import platform
import re
import sys
from urllib.parse import urlparse

from botocore import __version__ as __botocore_version__
from bs4 import __version__ as __bs4_version__
from bs4 import BeautifulSoup
from future import standard_library
import pytz
import requests
from tokendito import settings
from tokendito.__version__ import __version__
from tzlocal import get_localzone


standard_library.install_aliases()


def setup(args):
    """Parse command line arguments.

    :return: args parse object
    """
    parser = argparse.ArgumentParser(
        prog="tokendito", description="Gets a STS token to use with the AWS CLI"
    )
    parser.add_argument(
        "--version", "-v", action="store_true", help="Displays version and exit"
    )
    parser.add_argument(
        "--configure",
        "-c",
        action="store_true",
        help="Prompt user for " "configuration parameters",
    )
    parser.add_argument(
        "--username",
        "-u",
        type=to_unicode,
        dest="okta_username",
        help="username to login to Okta. You can "
        "also use the OKTA_USERNAME environment variable.",
    )
    parser.add_argument(
        "--password",
        "-p",
        type=to_unicode,
        dest="okta_password",
        help="password to login to Okta. You "
        "can also user the OKTA_PASSWORD environment variable.",
    )
    parser.add_argument(
        "--config-file",
        "-C",
        type=to_unicode,
        default=settings.config_file,
        help="Use an alternative configuration file",
    )
    parser.add_argument(
        "--okta-aws-app-url", "-ou", type=to_unicode, help="Okta App URL to use."
    )
    parser.add_argument(
        "--okta-profile",
        "-op",
        type=to_unicode,
        default=settings.okta_profile,
        help="Okta configuration profile to use.",
    )
    parser.add_argument(
        "--aws-region",
        "-r",
        type=to_unicode,
        help="Sets the AWS region for the profile",
    )
    parser.add_argument(
        "--aws-output",
        "-ao",
        type=to_unicode,
        help="Sets the AWS output type for the profile",
    )
    parser.add_argument(
        "--aws-profile",
        "-ap",
        type=to_unicode,
        help="Override AWS profile to save as in the credentials file.",
    )
    parser.add_argument(
        "--mfa-method", "-mm", type=to_unicode, help="Sets the MFA method"
    )
    parser.add_argument(
        "--mfa-response",
        "-mr",
        type=to_unicode,
        help="Sets the MFA response to a challenge",
    )
    parser.add_argument("--role-arn", "-R", type=to_unicode, help="Sets the IAM role")
    parser.add_argument(
        "--output-file", "-o", type=to_unicode, help="Log output to filename"
    )
    parser.add_argument(
        "--loglevel",
        "-l",
        type=lambda s: s.upper(),
        default="WARNING",
        choices=["DEBUG", "INFO", "WARN", "ERROR"],
        help="[DEBUG|INFO|WARN|ERROR], default loglevel is WARNING."
        " Note: DEBUG level may display credentials",
    )

    parsed_args = parser.parse_args(args)
    set_logging(parsed_args)
    logging.debug("Parse command line arguments [{}]".format(parsed_args))

    return parsed_args


def utc_to_local(utc_dt):
    """Convert UTC time into local time.

    :param:utc_str:datetime
    :return:local_time:string
    """
    local_time = utc_dt.replace(tzinfo=pytz.utc).astimezone(tz=get_localzone())
    local_time = local_time.strftime("%Y-%m-%d %H:%M:%S %Z")

    return local_time


def to_unicode(bytestring):
    """Convert a string into a Unicode compliant object.

    The `unicode()` method is only available in Python 2. Python 3
    generates a `NameError`, and the same string is returned unmodified.

    :param bytestring:
    :return: unicode-compliant string
    """
    if type(bytestring) == bytes:
        bytestring = bytestring.decode(settings.encoding)
    unicode_string = bytestring
    try:
        unicode_string = unicode(bytestring, settings.encoding)
    except (NameError, TypeError):
        # If a TypeError is raised, we are in Python 3, this is a no-op.
        pass
    return unicode_string


def create_directory(dir_name):
    """Create directories on the local machine."""
    if os.path.isdir(dir_name) is False:
        try:
            os.mkdir(dir_name)
        except OSError as error:
            logging.error(
                "Cannot continue creating directory '{}': {}".format(
                    settings.config_dir, error.strerror
                )
            )
            sys.exit(1)


def set_okta_username():
    """Set okta username in a constant settings variable.

    :return: okta_username

    """
    logging.debug("Set okta username in a constant settings variable.")

    if settings.okta_username == "":
        okta_username = input("Username: ")
        setattr(settings, "okta_username", to_unicode(okta_username))
        logging.debug("username set to {} interactively".format(settings.okta_username))

    return settings.okta_username


def set_okta_password():
    """Set okta password in a constant settings variable.

    :param args: command line arguments
    :return: okta_password

    """
    logging.debug("Set okta password in a constant settings variable.")

    while settings.okta_password == "":
        okta_password = getpass.getpass()
        setattr(settings, "okta_password", to_unicode(okta_password))

    logging.debug("password set interactively")
    return settings.okta_password


def set_logging(args):
    """Set logging level.

    :param args: Arguments provided by a user
    :return:

    """
    logger = logging.getLogger()
    logger.setLevel(args.loglevel)
    log_level_int = getattr(logging, args.loglevel)

    # increment boto logs to not print api keys
    logging.getLogger("botocore").setLevel(log_level_int + 10)

    log_format = (
        "%(levelname)s " "[%(filename)s:%(funcName)s():%(lineno)i]: %(message)s"
    )
    date_format = "%m/%d/%Y %I:%M:%S %p"

    formatter = logging.Formatter(log_format, date_format)

    if args.output_file:
        handler = logging.FileHandler(args.output_file)
    else:
        handler = logging.StreamHandler()

    handler.setFormatter(formatter)
    logger.addHandler(handler)


def select_role_arn(role_arns, saml_xml, saml_response_string):
    """Select the role user wants to pick.

    :param role_arns: IAM roles ARN list assigned for the user
    :param saml_xml: Decoded saml response from Okta
    :param saml_response_string: http response from saml assertion to AWS
    :return: User input index selected by the user, the arn of selected role

    """
    logging.debug("Select the role user wants to pick [{}]".format(role_arns))
    if settings.role_arn is None:
        selected_role = prompt_role_choices(role_arns, saml_xml, saml_response_string)
    elif settings.role_arn in role_arns:
        selected_role = settings.role_arn
    else:
        logging.error(
            "User provided rolename does not exist [{}]".format(settings.role_arn)
        )
        sys.exit(2)

    logging.debug("Selected role: [{}]".format(selected_role))

    return selected_role


def factor_type_info(factor_type, mfa_option):
    """Get factor info from okta reply.

    :param factor_type: mfa_method
    :param mfa_option: mfa_option
    :return: info about mfa_method
    """
    logging.debug("Choose factor info depending on factor type.")
    factor_info = "Not Presented"

    if factor_type in ["token", "token:software:totp", "token:hardware"]:
        factor_info = mfa_option.get("profile").get("credentialId")
    elif factor_type == "push":
        factor_info = mfa_option.get("profile").get("name")
    elif factor_type == "sms" or factor_type == "call":
        factor_info = mfa_option.get("profile").get("phoneNumber")
    elif factor_type == "webauthn":
        factor_info = mfa_option.get("profile").get("authenticatorName")
    elif factor_type in ["web", "u2f", "token:hotp"]:
        factor_info = mfa_option.get("vendorName")
    elif factor_type == "question":
        factor_info = mfa_option.get("profile").get("question")
    elif factor_type == "email":
        factor_info = mfa_option.get("profile").get("email")

    return factor_info


def mfa_option_info(mfa_option):
    """Build an optional string with the MFA factor information.

    :param mfa_option: dictionary with a single MFA response.
    :return: pre-formatted string with MFA factor info if available, None
             otherwise.
    """
    logging.debug("Building info for: {}".format(json.dumps(mfa_option)))

    if "factorType" in mfa_option:
        factor_type = mfa_option["factorType"]
        factor_info = factor_type_info(factor_type, mfa_option)

    if not factor_info:
        factor_info = "Not Presented"
    return factor_info


def select_preferred_mfa_index(
    mfa_options, factor_key="provider", subfactor_key="factorType"
):
    """Show all the MFA options to the users.

    :param mfa_options: List of available MFA options
    :return: MFA option selected index by the user from the output
    """
    logging.debug("Show all the MFA options to the users.")
    logging.debug(json.dumps(mfa_options))
    print("\nSelect your preferred MFA method and press Enter:")

    longest_index = len(str(len(mfa_options)))
    longest_factor_name = max([len(d[factor_key]) for d in mfa_options])
    longest_subfactor_name = max([len(d[subfactor_key]) for d in mfa_options])
    factor_info_indent = max([len(mfa_option_info(d)) for d in mfa_options])

    for (i, mfa_option) in enumerate(mfa_options):
        factor_id = mfa_option.get("id", "Not presented")
        factor_info = mfa_option_info(mfa_option)
        mfa_method = mfa_option.get(subfactor_key, "Not presented")
        provider = mfa_option.get(factor_key, "Not presented")
        print(
            "[{: >{}}]  {: <{}}  {: <{}} {: <{}} Id: {}".format(
                i,
                longest_index,
                provider,
                longest_factor_name,
                mfa_method,
                longest_subfactor_name,
                factor_info,
                factor_info_indent,
                factor_id,
            ),
        )

    user_input = collect_integer(len(mfa_options))

    return user_input


def prompt_role_choices(role_arns, saml_xml, saml_response_string):
    """Ask user to select role.

    :param role_arns: IAM Role list
    :return: user input of AWS Role
    """
    if len(role_arns) == 1:
        account_id = role_arns[0].split(":")[4]
        alias_table = {account_id: account_id}
    else:
        alias_table = get_account_aliases(saml_xml, saml_response_string)

    logging.debug("Ask user to select role")
    print("Please select one of the following:\n")

    longest_alias = max([len(d) for d in alias_table.values()])
    longest_index = len(str(len(role_arns)))
    sorted_role_arns = sorted(role_arns)

    for (i, arn) in enumerate(sorted_role_arns):
        padding_index = longest_index - len(str(i))
        account_alias = alias_table[arn.split(":")[4]]
        print(
            "[{}] {}{: <{}}    {}".format(
                i, padding_index * " ", account_alias, longest_alias, arn
            )
        )

    user_input = collect_integer(len(role_arns))
    selected_role = sorted_role_arns[user_input]
    logging.debug("Selected role [{}]".format(user_input))

    return selected_role


def print_selected_role(profile_name, expiration_time):
    """Print details about how to assume role.

    :param profile_name: AWS profile name
    :param expiration_time: Credentials expiration time
    :return:

    """
    expiration_time_local = utc_to_local(expiration_time)
    msg = (
        "\nGenerated profile '{}' in {}.\n"
        "\nUse profile to authenticate to AWS:\n\t"
        "aws --profile '{}' sts get-caller-identity"
        "\nOR\n\t"
        "export AWS_PROFILE='{}'\n\n"
        "Credentials are valid until {} ({})."
    ).format(
        profile_name,
        settings.aws_shared_credentials_file,
        profile_name,
        profile_name,
        expiration_time,
        expiration_time_local,
    )

    return print(msg)


def extract_arns(saml):
    """Extract arns from SAML decoded xml.

    :param saml: results saml decoded
    :return: Principle ARNs, Role ARNs
    """
    logging.debug("Decode response string as a SAML decoded value.")

    soup = BeautifulSoup(saml, "xml")
    arns = soup.find_all(text=re.compile("arn:aws:iam::"))
    if len(arns) == 0:
        logging.error("No IAM roles found in SAML response.")
        logging.debug(arns)
        sys.exit(2)

    roles_and_providers = {i.split(",")[1]: i.split(",")[0] for i in arns}

    logging.debug("Collected ARNs: {}".format(json.dumps(roles_and_providers)))

    return roles_and_providers


def validate_saml_response(html):
    """Parse html to validate that saml a saml response was returned."""
    soup = BeautifulSoup(html, "html.parser")

    xml = None
    for elem in soup.find_all("input", attrs={"name": "SAMLResponse"}):
        saml_base64 = elem.get("value")
        xml = codecs.decode(saml_base64.encode("ascii"), "base64").decode("utf-8")

    if xml is None:
        logging.error(
            "Invalid data detected in SAML response."
            " View the response with the DEBUG loglevel."
        )
        logging.debug(html)
        sys.exit(1)

    return xml


def validate_okta_aws_app_url(input_url=None):
    """Validate whether a given URL is a valid AWS app URL in Okta.

    :param input_url: string
    :return: bool. True if valid, False otherwise
    """
    logging.debug("Will try to match '{}' to a valid URL".format(input_url))

    url = urlparse(input_url)
    # Here, we could also check url.netloc against r'.*\.okta(preview)?\.com$'
    # but Okta allows the usage of custome URLs such as login.acme.com
    if (
        url.scheme == "https"
        and re.match(r"^/home/amazon_aws/\w{20}/\d{3}$", url.path) is not None
    ):
        return True

    logging.debug("{} does not look like a valid match.".format(url))
    return False


def get_account_aliases(saml_xml, saml_response_string):
    """Parse AWS SAML page for account aliases.

    :param saml_xml: Decoded saml response from Okta
    :param saml_response_string response from Okta with saml data:
    :return: mapping table of account ids to their aliases
    """
    soup = BeautifulSoup(saml_response_string, "html.parser")
    url = soup.find("form").get("action")

    encoded_xml = codecs.encode(saml_xml.encode("utf-8"), "base64")
    aws_response = None
    try:
        aws_response = requests.Session().post(url, data={"SAMLResponse": encoded_xml})
    except Exception as request_error:
        logging.error(
            "There was an error retrieving the AWS SAML page: \n{}".format(
                request_error
            )
        )
        logging.debug(json.dumps(aws_response))
        sys.exit(1)

    if "Account: " not in aws_response.text:
        logging.error("There were no accounts returned in the AWS SAML page.")
        logging.debug(json.dumps(aws_response.text))
        sys.exit(2)

    soup = BeautifulSoup(aws_response.text, "html.parser")
    account_names = soup.find_all(text=re.compile("Account:"))
    alias_table = {
        str(i.split(" ")[-1]).strip("()"): i.split(" ")[1] for i in account_names
    }

    return alias_table


def display_version():
    """Print program version and exit."""
    python_version = platform.python_version()
    (system, _, release, _, _, _) = platform.uname()
    print(
        "tokendito/{} Python/{} {}/{} botocore/{} bs4/{} requests/{}".format(
            __version__,
            python_version,
            system,
            release,
            __botocore_version__,
            __bs4_version__,
            requests.__version__,
        )
    )


def process_ini_file(file, profile):
    """Process options from a ConfigParser ini file.

    :param file: filename
    :param profile: profile to read
    :return: None
    """
    config = configparser.ConfigParser(default_section=settings.okta_profile)
    if config.read(file) == []:
        return

    try:
        for (key, val) in config.items(profile):
            if hasattr(settings, key):
                logging.debug("Set option {}={} from ini file".format(key, val))
                setattr(settings, key, val)
    except configparser.Error as err:
        logging.error("Could not load profile '{}': {}".format(profile, str(err)))
        sys.exit(2)


def process_arguments(args):
    """Process command-line arguments.

    :param args: argparse object
    :return: None
    """
    for (key, val) in vars(args).items():
        if hasattr(settings, key) and val is not None:
            logging.debug("Set option {}={} from command line".format(key, val))
            setattr(settings, key, val)


def process_environment():
    """Process environment variables.

    :return: None
    """
    for (key, val) in os.environ.items():
        key = key.lower()
        if hasattr(settings, key):
            logging.debug("Set option {}={} from environment".format(key, val))
            setattr(settings, key, os.getenv(key.upper()))


def process_okta_aws_app_url():
    """Process Okta app url.

    :param app_url: string with okta tile URL.
    :return: None.
    """
    if not validate_okta_aws_app_url(settings.okta_aws_app_url):
        logging.error(
            "Okta Application URL not found, or invalid. Please check "
            "your configuration and try again."
        )
        sys.exit(2)

    url = urlparse(settings.okta_aws_app_url)
    okta_org = "{}://{}".format(url.scheme, url.netloc)
    okta_aws_app_url = "{}{}".format(okta_org, url.path)
    setattr(settings, "okta_org", okta_org)
    setattr(settings, "okta_aws_app_url", okta_aws_app_url)


def user_configuration_input():
    """Obtain user input for the user.

    :return: (okta app url, organization username)
    """
    logging.debug("Obtain user input for the user.")
    url = ""
    username = ""
    config_details = []
    message = {
        "app_url": "\nOkta App URL. E.g https://acme.okta.com/home/"
        "amazon_aws/b07384d113edec49eaa6/123\n[none]: ",
        "username": "\nOrganization username. E.g jane.doe@acme.com" "\n[none]: ",
    }

    while url == "":
        user_data = to_unicode(input(message["app_url"]))
        user_data = user_data.strip()
        if validate_okta_aws_app_url(user_data):
            url = user_data
        else:
            print("Invalid input, try again.")
    config_details.append(url)

    while username == "":
        user_data = to_unicode(input(message["username"]))
        user_data = user_data.strip()
        if user_data != "":
            username = user_data
        else:
            print("Invalid input, try again.")
    config_details.append(username)

    return (config_details[0], config_details[1])


def update_configuration(okta_file, profile):
    """Update okta configuration file on local system.

    :param okta_file: Default configuration system file
    :param profile: profile of the okta user
    :return:
    """
    logging.debug("Update okta configuration file on local system.")

    config = configparser.RawConfigParser()

    create_directory(settings.config_dir)

    if os.path.isfile(okta_file):
        logging.debug("Read Okta config [{} {}]".format(okta_file, profile))
        config.read(okta_file, encoding=settings.encoding)
    if not config.has_section(profile):
        config.add_section(profile)
        logging.debug("Add section to Okta config [{}]".format(profile))

    (app_url, username) = user_configuration_input()

    url = urlparse(app_url.strip())
    okta_username = username.strip()

    okta_aws_app_url = "{}://{}{}".format(url.scheme, url.netloc, url.path)

    config.set(profile, "okta_aws_app_url", okta_aws_app_url)
    config.set(profile, "okta_username", okta_username)
    logging.debug("Config Okta [{}]".format(config))

    with open(okta_file, "w+", encoding=settings.encoding) as file:
        config.write(file)
        logging.debug("Write new section Okta config [{} {}]".format(okta_file, config))


def set_local_credentials(assume_role_response, role_name, aws_region, aws_output):
    """Write to local files to insert credentials.

    :param assume_role_response AWS AssumeRoleWithSaml response:
    :param role_name the name of the assumed role, used for local profile:
    :param aws_region configured region for aws credential profile:
    :param aws output configured datatype for aws cli output:
    """
    expiration_time = assume_role_response["Credentials"]["Expiration"]
    aws_access_key = assume_role_response["Credentials"]["AccessKeyId"]
    aws_secret_key = assume_role_response["Credentials"]["SecretAccessKey"]
    aws_session_token = assume_role_response["Credentials"]["SessionToken"]

    if settings.aws_profile is not None:
        role_name = settings.aws_profile

    update_aws_credentials(role_name, aws_access_key, aws_secret_key, aws_session_token)
    update_aws_config(role_name, aws_output, aws_region)

    print_selected_role(role_name, expiration_time)


def update_aws_credentials(profile, aws_access_key, aws_secret_key, aws_session_token):
    """Update AWS credentials in ~/.aws/credentials default file.

    :param profile: AWS profile name
    :param aws_access_key: AWS access key
    :param aws_secret_key: AWS secret access key
    :param aws_session_token: Session token
    """
    cred_file = settings.aws_shared_credentials_file
    cred_dir = os.path.dirname(cred_file)
    logging.debug("Update AWS credentials in: [{}]".format(cred_file))

    create_directory(cred_dir)

    config = configparser.RawConfigParser()
    if os.path.isfile(cred_file):
        config.read(cred_file, encoding=settings.encoding)
    if not config.has_section(profile):
        config.add_section(profile)
    config.set(profile, "aws_access_key_id", aws_access_key)
    config.set(profile, "aws_secret_access_key", aws_secret_key)
    config.set(profile, "aws_session_token", aws_session_token)
    with open(cred_file, "w+", encoding=settings.encoding) as file:
        config.write(file)


def update_aws_config(profile, output, region):
    """Update AWS config file in ~/.aws/config file.

    :param profile: tokendito profile
    :param output: aws output
    :param region: aws region
    :return:

    """
    config_file = settings.aws_config_file
    config_dir = os.path.dirname(config_file)
    logging.debug("Update AWS config to file: [{}]".format(config_file))

    create_directory(config_dir)

    # Prepend the word profile the the profile name
    profile = "profile {}".format(profile)
    config = configparser.RawConfigParser()
    if os.path.isfile(config_file):
        config.read(config_file, encoding=settings.encoding)
    if not config.has_section(profile):
        config.add_section(profile)
    config.set(profile, "output", output)
    config.set(profile, "region", region)

    with open(config_file, "w+", encoding=settings.encoding) as file:
        config.write(file)


def check_within_range(user_input, valid_range):
    """Validate the user input is within the range of the presented menu.

    :param user_input: integer-validated user input.
    :param valid_range: the valid range presented on the user's menu.
    :return range_validation: true or false
    """
    range_validation = False
    if int(user_input) in range(0, valid_range):
        range_validation = True
    else:
        logging.debug("Valid range is {}".format(valid_range))
        logging.error("Value is not in within the selection range.")
    return range_validation


def check_integer(value):
    """Validate integer.

    :param value: value to be validated.
    :return: True when the number is a positive integer, false otherwise.
    """
    integer_validation = False
    if str(value).isdigit():
        integer_validation = True
    else:
        logging.error("Please enter a valid integer.")

    return integer_validation


def validate_input(value, valid_range):
    """Validate user input is an integer and within menu range.

    :param value: user input
    :param valid_range: valid range based on how many menu options available to user.
    """
    integer_validation = check_integer(value)
    if integer_validation and valid_range:
        integer_validation = check_within_range(value, valid_range)
    return integer_validation


def get_input(prompt="-> "):
    """Collect user input for TOTP.

    :return user_input: raw from user.
    """
    user_input = to_unicode(input(prompt))
    logging.debug("User input [{}]".format(user_input))

    return user_input


def collect_integer(valid_range):
    """Collect input from user.

    Prompt the user for input. Validate it and cast to integer.

    :param valid_range: number of menu options available to user.
    :return user_input: validated, casted integer from user.
    """
    user_input = None
    while True:
        user_input = get_input()
        valid_input = validate_input(user_input, valid_range)
        logging.debug("User input validation status is {}".format(valid_input))
        if valid_input:
            user_input = int(user_input)
            break
    return user_input


def prepare_payload(**kwargs):
    """Prepare payload for the HTTP request header.

    :param kwargs: parameters to get together
    :return: payload for the http header
    """
    logging.debug("Prepare payload")

    payload_dict = {}
    if kwargs is not None:
        for key, value in list(kwargs.items()):
            payload_dict[key] = value

            if key != "password":
                logging.debug("Prepare payload [{} {}]".format(key, value))

    return payload_dict


def process_options(args):
    """Collect all user-specific credentials and config params."""
    if args.version:
        display_version()
        sys.exit(0)

    if args.configure:
        update_configuration(args.config_file, args.okta_profile)
        sys.exit(0)

    # 1: read ini file (if it exists)
    process_ini_file(args.config_file, args.okta_profile)
    # 2: override with args
    process_arguments(args)
    # 3: override with ENV
    process_environment()

    process_okta_aws_app_url()
    # Set username and password for Okta Authentication
    logging.debug("Set Okta credentials.")
    set_okta_username()
    set_okta_password()



# vim: set filetype=python ts=4 sw=4
# -*- coding: utf-8 -*-
"""tokendito version."""
__version__ = "1.2.0"
__title__ = "tokendito"
__description__ = "Get AWS STS tokens from Okta SSO"
__long_description_content_type__ = "text/x-rst"
__url__ = "https://github.com/dowjones/tokendito"
__author__ = "tokendito"
__author_email__ = "tokendito@dowjones.com"
__license__ = "Apache 2.0"



# vim: set filetype=python ts=4 sw=4
# -*- coding: utf-8 -*-
"""
This module handles the all Okta operations.

1. Okta authentication
2. Update Okta Config File

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import logging
import sys
import time


from future import standard_library
import requests
from tokendito import duo_helpers
from tokendito import helpers
from tokendito import settings


standard_library.install_aliases()


def okta_verify_api_method(mfa_challenge_url, payload, headers=None):
    """Okta MFA authentication.

    :param mfa_challenge_url: MFA challenge url
    :param payload: JSON Payload
    :param headers: Headers of the request
    :return: Okta authentication response
    """
    try:
        if headers:
            response = requests.request(
                "POST", mfa_challenge_url, data=json.dumps(payload), headers=headers
            )
        else:
            response = requests.request("POST", mfa_challenge_url, data=payload)
    except Exception as request_error:
        logging.error(
            "There was an error connecting to Okta: \n{}".format(request_error)
        )
        sys.exit(1)

    logging.debug("Okta authentication response: \n{}".format(response))

    try:
        response = response.json()
    except ValueError:
        logging.debug("Received type of response: {}".format(type(response.text)))
        response = response.text

    if "errorCode" in response:
        login_error_code_parser(response["errorCode"], settings.okta_status_dict)
        sys.exit(1)

    return response


def login_error_code_parser(status=None, status_dict=settings.okta_status_dict):
    """Status code parsing.

    param status: Response status
    return message: status message
    """
    if status in status_dict.keys():
        message = "Okta auth failed: {}".format(status_dict[status])
    else:
        message = (
            "Okta auth failed: {}. Please verify your settings and try again.".format(
                status
            )
        )
    logging.error(message)
    logging.debug("Parsing error [{}] ".format(message))
    return message


def user_session_token(primary_auth, headers):
    """Get session_token.

    param headers: Headers of the request
    param primary_auth: Primary authentication
    return session_token: Session Token from JSON response
    """
    try:
        status = primary_auth.get("errorCode", primary_auth["status"])
    except KeyError:
        logging.debug("Error parsing response: {}".format(json.dumps(primary_auth)))
        logging.error("Okta auth failed: unknown status")
        sys.exit(1)
    if status == "SUCCESS":
        session_token = primary_auth["sessionToken"]
    elif status == "MFA_REQUIRED":
        session_token = user_mfa_challenge(headers, primary_auth)
    else:
        login_error_code_parser(status, settings.okta_status_dict)
        sys.exit(1)
    return session_token


def authenticate_user(okta_url, okta_username, okta_password):
    """Authenticate user with okta credential.

    :param okta_url: company specific URL of the okta
    :param okta_username: okta username
    :param okta_password: okta password
    :return: MFA session options

    """
    logging.debug(
        "Authenticate user with okta credential [{} user {}]".format(
            okta_url, okta_username
        )
    )
    headers = {"content-type": "application/json", "accept": "application/json"}
    payload = helpers.prepare_payload(username=okta_username, password=okta_password)

    primary_auth = okta_verify_api_method(
        "{}/api/v1/authn".format(okta_url), payload, headers
    )
    logging.debug("Authenticate Okta header [{}] ".format(headers))

    session_token = user_session_token(primary_auth, headers)

    logging.info("User has been succesfully authenticated.")
    return session_token


def mfa_provider_type(
    mfa_provider,
    selected_factor,
    mfa_challenge_url,
    primary_auth,
    selected_mfa_option,
    headers,
    payload,
):
    """Receive session key.

    :param mfa_provider: MFA provider
    :param selected_factor: Selected MFA factor
    :param mfa_challenge_url: MFA challenge url
    :param primary_auth: Primary authentication
    :param selected_mfa_option: Selected MFA option
    :return: session_key

    """
    if mfa_provider == "duo":
        payload, headers, callback_url = duo_helpers.authenticate_duo(selected_factor)
        okta_verify_api_method(callback_url, payload)
        payload.pop("id", "sig_response")
        mfa_verify = okta_verify_api_method(mfa_challenge_url, payload, headers)
    elif mfa_provider == "okta" or mfa_provider == "google":
        mfa_verify = user_mfa_options(
            selected_mfa_option, headers, mfa_challenge_url, payload, primary_auth
        )
    else:
        logging.error(
            "Sorry, the MFA provider '{}' is not yet supported."
            " Please retry with another option.".format(mfa_provider)
        )
        exit(1)
    return mfa_verify["sessionToken"]


def user_mfa_index(preset_mfa, available_mfas, mfa_options):
    """Get mfa method index in request.

    :param preset_mfa: preset mfa method from settings
    :param available_mfas: available mfa method ids
    :param mfa_options: available mfa methods
    """
    logging.debug("Get mfa method index in request.")
    if preset_mfa is not None and preset_mfa in available_mfas:
        mfa_index = available_mfas.index(preset_mfa)
    else:
        mfa_index = helpers.select_preferred_mfa_index(mfa_options)

    return mfa_index


def user_mfa_challenge(headers, primary_auth):
    """Handle user mfa challenges.

    :param headers: headers what needs to be sent to api
    :param primary_auth: primary authentication
    :return: Okta MFA Session token after the successful entry of the code
    """
    logging.debug("Handle user MFA challenges")

    mfa_options = primary_auth["_embedded"]["factors"]
    preset_mfa = settings.mfa_method

    available_mfas = [d["factorType"] for d in mfa_options]

    if available_mfas.count(preset_mfa) > 1:
        mfa_method = settings.mfa_method
        mfa_index = available_mfas.index(preset_mfa)
        provider = mfa_options[mfa_index]["provider"]
        mfa_id = mfa_options[mfa_index]["id"]

        logging.warning(
            "\n\nMore than one method found with {}.\n"
            "Defaulting to {} - {} - Id: {}.\n"
            "This functionality will be deprecated in"
            "the next major release.\n".format(mfa_method, provider, mfa_method, mfa_id)
        )

    mfa_index = user_mfa_index(preset_mfa, available_mfas, mfa_options)

    # time to challenge the mfa option
    selected_mfa_option = mfa_options[mfa_index]
    logging.debug("Selected MFA is [{}]".format(selected_mfa_option))

    mfa_challenge_url = selected_mfa_option["_links"]["verify"]["href"]

    payload = helpers.prepare_payload(
        stateToken=primary_auth["stateToken"],
        factorType=selected_mfa_option["factorType"],
        provider=selected_mfa_option["provider"],
        profile=selected_mfa_option["profile"],
    )
    selected_factor = okta_verify_api_method(mfa_challenge_url, payload, headers)

    mfa_provider = selected_factor["_embedded"]["factor"]["provider"].lower()
    logging.debug(
        "MFA Challenge URL: [{}] headers: {}".format(mfa_challenge_url, headers)
    )
    mfa_session_token = mfa_provider_type(
        mfa_provider,
        selected_factor,
        mfa_challenge_url,
        primary_auth,
        selected_mfa_option,
        headers,
        payload,
    )

    return mfa_session_token


def user_mfa_options(
    selected_mfa_option, headers, mfa_challenge_url, payload, primary_auth
):
    """Handle user mfa options.

    :param selected_mfa_option: Selected MFA option (SMS, push, etc)
    :param headers: headers
    :param mfa_challenge_url: MFA challenge URL
    :param payload: payload
    :param primary_auth: Primary authentication method
    :return: payload data

    """
    logging.debug("Handle user MFA options")

    logging.debug(
        "User MFA options selected: [{}]".format(selected_mfa_option["factorType"])
    )
    if selected_mfa_option["factorType"] == "push":
        return push_approval(headers, mfa_challenge_url, payload)

    if settings.mfa_response is None:
        logging.debug("Getting verification code from user.")
        print("Type verification code and press Enter")
        settings.mfa_response = helpers.get_input()

    # time to verify the mfa method
    payload = helpers.prepare_payload(
        stateToken=primary_auth["stateToken"], passCode=settings.mfa_response
    )
    mfa_verify = okta_verify_api_method(mfa_challenge_url, payload, headers)
    logging.debug("mfa_verify [{}]".format(json.dumps(mfa_verify)))

    return mfa_verify


def push_approval(headers, mfa_challenge_url, payload):
    """Handle push approval from the user.

    :param headers: HTTP headers sent to API call
    :param mfa_challenge_url: MFA challenge url
    :param payload: payload which needs to be sent
    :return: Session Token if succeeded or terminates if user wait goes 5 min

    """
    logging.debug(
        "Handle push approval from the user [{}] [{}]".format(
            headers, mfa_challenge_url
        )
    )

    print("Waiting for an approval from device...")
    mfa_status = "WAITING"

    while mfa_status == "WAITING":
        mfa_verify = okta_verify_api_method(mfa_challenge_url, payload, headers)

        logging.debug("MFA Response:\n{}".format(json.dumps(mfa_verify)))

        if "factorResult" in mfa_verify:
            mfa_status = mfa_verify["factorResult"]
        elif mfa_verify["status"] == "SUCCESS":
            break
        else:
            logging.error("There was an error getting your MFA status.")
            logging.debug(mfa_verify)
            if "status" in mfa_verify:
                logging.error("Exiting due to error: {}".format(mfa_verify["status"]))
            sys.exit(1)

        if mfa_status == "REJECTED":
            logging.error("The Okta Verify push has been denied. Please retry later.")
            sys.exit(2)
        elif mfa_status == "TIMEOUT":
            logging.error("Device approval window has expired.")
            sys.exit(2)

        time.sleep(2)

    return mfa_verify



#!/usr/bin/env python
# vim: set filetype=python ts=4 sw=4
# -*- coding: utf-8 -*-
"""tokendito cli entry point."""
from __future__ import absolute_import, division, print_function, unicode_literals

import sys

from future import standard_library

standard_library.install_aliases()


def main(args=None):  # needed for console script
    """Packge entry point."""
    if __package__ is None:
        import os.path

        path = os.path.dirname(os.path.dirname(__file__))
        sys.path[0:0] = [path]
    from tokendito.tool import cli

    return cli(args)


if __name__ == "__main__":
    try:
        sys.exit(main(sys.argv[1:]))
    except KeyboardInterrupt:
        print("\nInterrupted")
        sys.exit(1)



#!/usr/bin/env python
# vim: set filetype=python ts=4 sw=4
# -*- coding: utf-8 -*-
"""`tokendito` is in Github: <https://github.com/dowjones/tokendito>_."""

from codecs import open
import datetime
import os
import sys

from setuptools import find_packages, setup


here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.rst"), encoding=sys.stdin.encoding) as f:
    long_description = f.read()

with open("requirements.txt") as f:
    required = f.read().splitlines()

about = {}
with open(os.path.join(here, "tokendito", "__version__.py"), "r") as f:
    exec(f.read(), about)

if "DEVBUILD" in os.environ:
    now = datetime.datetime.now()
    about["__version__"] = about["__version__"] + ".dev" + now.strftime("%Y%m%d%H%M%S")

setup(
    name="tokendito",
    version=about["__version__"],
    description=about["__description__"],
    long_description=long_description,
    long_description_content_type=about["__long_description_content_type__"],
    url=about["__url__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
        "Environment :: Console",
        "Programming Language :: Python",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords=["okta", "aws", "sts"],
    packages=find_packages(exclude=["contrib", "docs", "tests", ".tox"]),
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
    license=about["__license__"],
    zip_safe=False,
    install_requires=[required],
    entry_points={
        "console_scripts": ["tokendito=tokendito.__main__:main"],
    },
    # $ pip install -e . [dev,test]
)



# vim: set filetype=python ts=4 sw=4
# -*- coding: utf-8 -*-
"""This module handles Duo operations."""
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import logging
import sys
import time
from urllib.parse import unquote, urlparse

from bs4 import BeautifulSoup
from future import standard_library
import requests
from tokendito import helpers
from tokendito import settings

standard_library.install_aliases()


def prepare_duo_info(selected_okta_factor):
    """Aggregate most of the parameters needed throughout the Duo authentication process.

    :param selected_okta_factor: dict response describing Duo factor in Okta.
    :return duo_info: dict of parameters for Duo
    """
    duo_info = {}
    okta_factor = selected_okta_factor["_embedded"]["factor"]["_embedded"][
        "verification"
    ]
    duo_info["okta_factor"] = okta_factor
    duo_info["factor_id"] = selected_okta_factor["_embedded"]["factor"]["id"]

    duo_info["state_token"] = selected_okta_factor["stateToken"]
    duo_info["okta_callback_url"] = okta_factor["_links"]["complete"]["href"]
    duo_info["tx"] = okta_factor["signature"].split(":")[0]
    duo_info["app_sig"] = okta_factor["signature"].split(":")[1]
    duo_info["parent"] = "{}/signin/verify/duo/web".format(settings.okta_org)
    duo_info["host"] = okta_factor["host"]
    duo_info["sid"] = ""

    version = okta_factor["_links"]["script"]["href"].split("-v")[1]
    duo_info["version"] = version.strip(".js")

    return duo_info


def duo_api_post(url, params={}, headers={}, payload={}):
    """Error handling and response parsing wrapper for Duo POSTs.

    :param url: The URL being connected to.
    :param params: URL query parameters.
    :param headers: Request headers.
    :param payload: Request body.
    :return response: Response to the API request.
    """
    try:
        response = requests.request(
            "POST", url, params=params, headers=headers, data=payload
        )
    except Exception as request_issue:
        logging.error(
            "There was an error connecting to the Duo API: \n{}".format(request_issue)
        )
        sys.exit(1)

    json_message = None
    try:
        json_message = response.json()
    except ValueError:
        logging.debug("Non-json response from Duo API: \n{}".format(response))

    if response.status_code != 200:
        logging.critical(
            "Your Duo authentication has failed with status {}.".format(
                response.status_code
            )
        )
        if json_message and json_message["stat"].lower() != "ok":
            logging.critical(
                "\n{}, {}".format(response.status_code, json_message["message"])
            )
        else:
            logging.critical(
                "Please re-run the program with parameter"
                ' "--loglevel debug" to see more information.'
            )
        sys.exit(2)

    return response


def get_duo_sid(duo_info):
    """Perform the initial Duo authentication request to obtain the SID.

    The SID is referenced throughout the authentication process for Duo.

    :param duo_info: dict response describing Duo factor in Okta.
    :return: duo_info with added SID.
    :return: duo_auth_response, contains html content listing available factors.
    """
    params = helpers.prepare_payload(
        tx=duo_info["tx"], v=duo_info["version"], parent=duo_info["parent"]
    )

    url = "https://{}/frame/web/v1/auth".format(duo_info["host"])
    logging.info(
        "Calling Duo {} with params {}".format(urlparse(url).path, params.keys())
    )
    duo_auth_response = duo_api_post(url, params=params)

    try:
        duo_auth_redirect = urlparse("{}".format(unquote(duo_auth_response.url))).query
        duo_info["sid"] = duo_auth_redirect.strip("sid=")
    except Exception as sid_error:
        logging.error(
            "There was an error getting your SID."
            "Please try again. \n{}".format(sid_error)
        )

    return duo_info, duo_auth_response


def get_duo_devices(duo_auth):
    """Parse Duo auth response to extract user's MFA options.

    The /frame/web/v1/auth API returns an html page that lists
    devices and their mfa options for the user logging in.
    The return data type (list of dicts) is intended to allow us to
    do printout padding and indexing when interacting with the end user.

    :param duo_auth: contains html content listing available factors.
    :return factor_options: list of dict objects describing each MFA option.
    """
    soup = BeautifulSoup(duo_auth.content, "html.parser")

    device_soup = soup.find("select", {"name": "device"}).findAll("option")
    devices = ["{} - {}".format(d["value"], d.text) for d in device_soup]
    if not devices:
        logging.error("Please configure devices for your Duo MFA and retry.")
        sys.exit(2)

    factor_options = []
    for device in devices:
        options = soup.find("fieldset", {"data-device-index": device.split(" - ")[0]})
        factors = options.findAll("input", {"name": "factor"})
        for factor in factors:
            factor_option = {}
            factor_option["device"] = device
            factor_option["factor"] = factor["value"]
            factor_options.append(factor_option)
    return factor_options


def parse_duo_mfa_challenge(mfa_challenge):
    """Gracefully parse Duo MFA challenge response.

    :param mfa_challenge: Duo API response for MFA challenge.
    :return txid: Duo transaction ID.
    """
    try:
        mfa_challenge = mfa_challenge.json()
        mfa_status = mfa_challenge["stat"]
        txid = mfa_challenge["response"]["txid"]
    except ValueError as value_error:
        logging.error(
            "The Duo API returned a non-json response: \n{}".format(value_error)
        )
        sys.exit(1)
    except KeyError as key_error:
        logging.error(
            "The Duo API response is missing a required parameter: \n{}".format(
                key_error
            )
        )
        print(json.dumps(mfa_challenge))
        sys.exit(1)

    if mfa_status == "fail":
        logging.error(
            "Your Duo authentication has failed: \n{}".format(mfa_challenge["message"])
        )
        sys.exit(1)
    return txid


def duo_mfa_challenge(duo_info, mfa_option, passcode):
    """Poke Duo to challenge the selected factor.

    After the user has selected their device and factor of choice,
    tell Duo to send a challenge. This is where the end user will receive
    a phone call or push.

    :param duo_info: dict of parameters for Duo
    :param mfa_option: the user's selected second factor.
    :return txid: Duo transaction ID used to track this auth attempt.
    """
    url = "https://{}/frame/prompt".format(duo_info["host"])
    device = mfa_option["device"].split(" - ")[0]
    mfa_data = helpers.prepare_payload(
        factor=mfa_option["factor"],
        device=device,
        sid=duo_info["sid"],
        out_of_date=False,
        days_out_of_date=0,
        days_to_block=None,
    )
    mfa_data["async"] = True  # async is a reserved keyword
    if passcode:
        mfa_data["passcode"] = passcode
    mfa_challenge = duo_api_post(url, payload=mfa_data)
    txid = parse_duo_mfa_challenge(mfa_challenge)

    logging.debug("Sent MFA Challenge and obtained Duo transaction ID.")
    return txid


def get_mfa_response(mfa_result):
    """Extract json from mfa response.

    :param mfa_result: raw response from mfa api
    :return verify_mfa: json response from mfa api
    """
    try:
        verify_mfa = mfa_result.json()["response"]
    except Exception as parse_error:
        logging.error(
            "There was an error parsing the mfa challenge result: \n{}".format(
                parse_error
            )
        )
        sys.exit(1)
    return verify_mfa


def parse_challenge(verify_mfa, challenge_result):
    """Parse the challenge response.

    :param mfa_result: response from MFA challenge status request
    :return challenge status: status of MFA challenge
    :return challenge reason: additional info about challenge status
    """
    challenge_reason = None

    if "status" in verify_mfa:
        print(verify_mfa["status"])

    if "reason" in verify_mfa:
        challenge_reason = verify_mfa["reason"]

    if "result" in verify_mfa:
        logging.info("Result received: {}".format(verify_mfa["result"]))
        challenge_result = verify_mfa["result"].lower()

    logging.debug("Challenge result is {}".format(challenge_result))
    return challenge_result, challenge_reason


def duo_mfa_verify(duo_info, txid):
    """Verify MFA challenge completion.

    After the user has received the MFA challenge, query the Duo API
    until the challenge is completed.

    :param duo_info: dict of parameters for Duo.
    :param mfa_option: the user's selected second factor.
    :return txid: Duo transaction ID used to track this auth attempt.
    """
    url = "https://{}/frame/status".format(duo_info["host"])
    challenged_mfa = helpers.prepare_payload(txid=txid, sid=duo_info["sid"])
    challenge_result = None

    while True:
        logging.debug("Waiting for MFA challenge response")
        mfa_result = duo_api_post(url, payload=challenged_mfa)
        verify_mfa = get_mfa_response(mfa_result)
        challenge_result, challenge_reason = parse_challenge(
            verify_mfa, challenge_result
        )

        if challenge_result is None:
            continue
        elif challenge_result == "success":
            logging.debug("Successful MFA challenge received")
            break
        elif challenge_result == "failure":
            logging.critical(
                "MFA challenge has failed:"
                " {}. Please try again.".format(challenge_reason)
            )
            sys.exit(2)
        else:
            logging.debug(
                "MFA challenge result: {}"
                "Reason: {}\n\n".format(challenge_result, challenge_reason)
            )
        time.sleep(1)

    return verify_mfa


def duo_factor_callback(duo_info, verify_mfa):
    """Inform factor callback api of successful challenge.

    This request seems to inform this factor's callback url
    that the challenge process has been completed.

    :param duo_info: dict of parameters for Duo.
    :param verify_mfa: verified mfa challenge response from status api.
    :return sig_response: required to sign final Duo callback request.
    """
    factor_callback_url = "https://{}{}".format(
        duo_info["host"], verify_mfa["result_url"]
    )
    factor_callback = duo_api_post(
        factor_callback_url, payload={"sid": duo_info["sid"]}
    )

    try:
        sig_response = "{}:{}".format(
            factor_callback.json()["response"]["cookie"], duo_info["app_sig"]
        )
    except Exception as sig_error:
        logging.error(
            "There was an error getting your"
            " application signature from Duo: \n{}".format(json.dumps(sig_error))
        )

    logging.debug("Completed factor callback.")
    return sig_response


def set_passcode(mfa_option):
    """Set totp passcode.

    If the user has selected the passcode option, collect their TOTP.

    :param mfa_option: selected factor
    :return passcode: passcode value from user
    """
    passcode = None
    if mfa_option["factor"].lower() == "passcode":
        print("Type your TOTP and press Enter")
        passcode = helpers.get_input()
    return passcode


def authenticate_duo(selected_okta_factor):
    """Accomplish MFA via Duo.

    This is the main function that coordinates the Duo
    multifactor fetching, presentation, selection, challenge,
    and verification until making an Okta callback.

    :param selected_okta_factor: Duo factor information retrieved from Okta.
    :return payload: required payload for Okta callback
    :return headers: required headers for Okta callback
    """
    try:
        duo_info = prepare_duo_info(selected_okta_factor)
    except KeyError as missing_key:
        logging.error(
            "There was an issue parsing the Okta factor."
            " Please try again. \n{}".format(missing_key)
        )
        sys.exit(1)

    # Collect devices, factors, auth params for Duo
    duo_info, duo_auth_response = get_duo_sid(duo_info)
    factor_options = get_duo_devices(duo_auth_response)
    mfa_index = helpers.select_preferred_mfa_index(
        factor_options, factor_key="factor", subfactor_key="device"
    )

    mfa_option = factor_options[mfa_index]
    logging.debug("Selected MFA is [{}]".format(mfa_option))
    passcode = set_passcode(mfa_option)

    txid = duo_mfa_challenge(duo_info, mfa_option, passcode)
    verify_mfa = duo_mfa_verify(duo_info, txid)

    # Make factor callback to Duo
    sig_response = duo_factor_callback(duo_info, verify_mfa)

    # Prepare for Okta callback
    payload = helpers.prepare_payload(
        id=duo_info["factor_id"],
        sig_response=sig_response,
        stateToken=duo_info["state_token"],
    )
    headers = {}
    headers["content-type"] = "application/json"
    headers["accept"] = "application/json"

    return payload, headers, duo_info["okta_callback_url"]



# vim: set filetype=python ts=4 sw=4
# -*- coding: utf-8 -*-
"""This module handles the all aws workflow operations.

Tasks include:
1. Aws Authentication with SAML
2. Updating the AWS Credentials
3. Updating the AWS Config

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import codecs
import logging
import sys

import boto3
from botocore import UNSIGNED
from botocore.client import Config
from botocore.exceptions import ClientError
from future import standard_library
import requests
from tokendito import helpers


standard_library.install_aliases()


def authenticate_to_roles(secret_session_token, url):
    """Authenticate AWS user with saml.

    :param secret_session_token: secret session token
    :param url: url of the AWS account
    :return: response text

    """
    payload = {"onetimetoken": secret_session_token}
    logging.debug("Authenticate AWS user with SAML URL [{}:{}]".format(url, payload))
    try:
        response = requests.get(url, params=payload)
        saml_response_string = response.text
        if response.status_code == 400 or response.status_code == 401:
            errmsg = "Invalid Credentials."
            logging.critical(
                "{}\nExiting with code:{}".format(errmsg, response.status_code)
            )
            sys.exit(2)
        elif response.status_code == 404:
            errmsg = "Invalid Okta application URL. Please verify your configuration."
            logging.critical("{}".format(errmsg))
            sys.exit(2)
        elif response.status_code >= 500 and response.status_code < 504:
            errmsg = "Unable to establish connection with Okta. Verify Okta Org URL and try again."
            logging.critical(
                "{}\nExiting with code:{}".format(errmsg, response.status_code)
            )
            sys.exit(2)
        elif response.status_code != 200:
            logging.critical("Exiting with code:{}".format(response.status_code))
            logging.debug(saml_response_string)
            sys.exit(2)

    except Exception as error:
        errmsg = "Okta auth failed:\n{}".format(error)
        logging.critical(errmsg)
        sys.exit(1)

    saml_xml = helpers.validate_saml_response(saml_response_string)

    return saml_response_string, saml_xml


def assume_role(role_arn, provider_arn, saml):
    """Return AssumeRoleWithSaml API response.

    :param role_arn: IAM role arn to assume
    :param provider_arn: ARN of saml-provider resource
    :param saml: decoded saml response from okta
    :return: AssumeRoleWithSaml API response

    """
    default_error = (
        "\nUnable to assume role with the following details:\n"
        "- Role ARN: {}\n"
        "- Error: {}\n"
    )

    encoded_xml = codecs.encode(saml.encode("utf-8"), "base64")

    # Attempt to assume a role with the following durations:
    # 12h, 8h, 6h, 4h, 2h, 1h, 30m, 15m
    session_times = [43200, 28800, 21600, 14400, 7200, 3600, 1800, 900, "exit"]
    for duration in session_times:
        if duration == "exit":
            logging.critical(
                default_error.format(
                    role_arn,
                    "IAM role session time is not within set: {}".format(
                        session_times[:-1]
                    ),
                )
            )
            sys.exit(2)

        assume_role_response = handle_assume_role(
            role_arn, provider_arn, encoded_xml, duration, default_error
        )
        if "Credentials" in assume_role_response:
            break

    return assume_role_response


def handle_assume_role(role_arn, provider_arn, encoded_xml, duration, default_error):
    """Handle assume role with saml.

    :param role_arn: IAM role arn to assume
    :param provider_arn: ARN of saml-provider resource
    :param saml: decoded saml response from okta
    :return: AssumeRoleWithSaml API responses
    """
    logging.debug("Attempting session time [{}]".format(duration))
    client = boto3.client("sts", config=Config(signature_version=UNSIGNED))
    try:
        assume_role_response = client.assume_role_with_saml(
            RoleArn=role_arn,
            PrincipalArn=provider_arn,
            SAMLAssertion=encoded_xml.decode(),
            DurationSeconds=duration,
        )
        # Client Exceptions
    except ClientError as error:
        if error.response["Error"]["Code"] == "ValidationError":
            logging.info(
                "AssumeRoleWithSaml failed with {} for duration {}".format(
                    error.response["Error"]["Code"], duration
                )
            )
            assume_role_response = "continue"
        elif error.response["Error"]["Code"] == "AccessDenied":
            errmsg = "Error assuming intermediate {} SAML role".format(provider_arn)
            logging.critical(errmsg)
            sys.exit(2)
        else:
            logging.critical(default_error.format(role_arn, str(error)))
            sys.exit(1)
        # Service Exceptions
    except Exception as error:
        logging.critical(default_error.format(role_arn, str(error)))
        sys.exit(1)

    return assume_role_response


def ensure_keys_work(assume_role_response):
    """Validate the temporary AWS credentials.

    :param aws_access_key: AWS access key
    :param aws_secret_key: AWS secret key
    :param aws_session_token: AWS session token
    :return:

    """
    logging.debug("Validate the temporary AWS credentials")

    aws_access_key = assume_role_response["Credentials"]["AccessKeyId"]
    aws_secret_key = assume_role_response["Credentials"]["SecretAccessKey"]
    aws_session_token = assume_role_response["Credentials"]["SessionToken"]

    try:
        client = boto3.client(
            "sts",
            aws_access_key_id=aws_access_key,
            aws_secret_access_key=aws_secret_key,
            aws_session_token=aws_session_token,
        )
        client.get_caller_identity()
    except Exception as auth_error:
        logging.critical(
            "There was an error authenticating your keys with AWS: {}".format(
                auth_error
            )
        )
        sys.exit(1)


def select_assumeable_role(saml_response_string, saml):
    """Select the role to perform the AssumeRoleWithSaml.

    :param saml_response_string response from Okta with saml data:
    :param saml decoded saml response from Okta:
    :return AWS AssumeRoleWithSaml response, role name:
    """
    roles_and_providers = helpers.extract_arns(saml)
    role_arn = helpers.select_role_arn(
        list(roles_and_providers.keys()), saml, saml_response_string
    )
    role_name = role_arn.split("/")[-1]

    assume_role_response = assume_role(role_arn, roles_and_providers[role_arn], saml)

    return assume_role_response, role_name



# vim: set filetype=python ts=4 sw=4
# -*- coding: utf-8 -*-
"""This module retrieves AWS credentials after authenticating with Okta."""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from future import standard_library
from tokendito import aws_helpers
from tokendito import helpers
from tokendito import okta_helpers
from tokendito import settings

standard_library.install_aliases()


def cli(args):
    """Tokendito retrieves AWS credentials after authenticating with Okta."""
    # Set some required initial values
    args = helpers.setup(args)

    logging.debug("tokendito retrieves AWS credentials after authenticating with Okta.")

    # Collect and organize user specific information
    helpers.process_options(args)

    # Authenticate okta and AWS also use assumerole to assign the role
    logging.debug("Authenticate user with Okta and AWS.")

    secret_session_token = okta_helpers.authenticate_user(
        settings.okta_org, settings.okta_username, settings.okta_password
    )

    saml_response_string, saml_xml = aws_helpers.authenticate_to_roles(
        secret_session_token, settings.okta_aws_app_url
    )

    assume_role_response, role_name = aws_helpers.select_assumeable_role(
        saml_response_string, saml_xml
    )

    aws_helpers.ensure_keys_work(assume_role_response)

    helpers.set_local_credentials(
        assume_role_response, role_name, settings.aws_region, settings.aws_output
    )
